#!/usr/bin/env python3
'''Addon functions that need additional dependencies to work.

To also install additional dependencies, use::

    pip install eminus[addons]
'''
from .fods import get_fods, remove_core_fods
from .viewer import view_grid, view_mol

__all__ = ['get_fods', 'remove_core_fods', 'view_grid', 'view_mol']



#!/usr/bin/env python3
'''parametrizations of exchange-correlation functionals.'''
import numpy as np

from .logger import log


def get_xc(xc, n, spinpol=False):
    '''Handle and get exchange-correlation functionals.

    Args:
        xc (str): Exchange and correlation identifier, separated by a comma.
        n (ndarray): Real-space electronic density.

    Keyword Args:
        spinpol (bool): Choose if a spin-polarized exchange-correlation functional will be used.

    Returns:
        tuple[ndarray, ndarray]: Exchange-correlation energy density and potential.
    '''
    xc_map = {
        'lda': 'lda_slater_x',
        'pw': 'lda_pw_c',
        'vwn': 'lda_vwn_c'
    }
    exch, corr = xc.split(',')

    # Only use non-zero values of the density
    n_nz = n[np.nonzero(n)]

    # Only import libxc interface if necessary
    if 'libxc' in xc:
        from .addons.libxc import libxc_functional

    # Handle exchange part
    if 'libxc' in exch:
        exch = exch.split(':')[1]
        ex, vx = libxc_functional(exch, n_nz, spinpol)
    else:
        # If the desired functional is not implemented use a mock functional
        try:
            f_exch = xc_map[exch]
        except KeyError:
            f_exch = 'mock_xc'
            log.warning('Use a mock functional for the exchange part.')
        if spinpol:
            f_exch = f'{f_exch}_spin'
        # FIXME: In spin-polarized calculations zeta is normally not one, only when coming from
        #        spin-unpolarized calculations
        ex, vx = eval(f_exch)(n_nz, zeta=np.ones_like(n_nz))

    # Handle correlation part
    if 'libxc' in corr:
        corr = corr.split(':')[1]
        ec, vc = libxc_functional(corr, n_nz, spinpol)
    else:
        # If the desired functional is not implemented use a mock functional
        try:
            f_corr = xc_map[corr]
        except KeyError:
            f_corr = 'mock_xc'
            log.warning('Use a mock functional for the correlation part.')
        if spinpol:
            f_corr = f'{f_corr}_spin'
        # FIXME: In spin-polarized calculations zeta is normally not one, only when coming from
        #        spin-unpolarized calculations
        ec, vc = eval(f_corr)(n_nz, zeta=np.ones_like(n_nz))

    # Use zero as the result for zero values
    exc = np.zeros_like(n)
    exc[np.nonzero(n)] = ex + ec
    vxc = np.zeros_like(n)
    vxc[np.nonzero(n)] = vx + vc
    return exc, vxc


def mock_xc(n, **kwargs):
    '''Mock exchange-correlation functional with no effect.

    Args:
        n (ndarray): Real-space electronic density.

    Returns:
        tuple[ndarray, ndarray]: Mock exchange-correlation energy density and potential.
    '''
    zeros = np.zeros_like(n)
    return zeros, zeros


mock_xc_spin = mock_xc


# Adapted from https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/XC_funcs/XC_x_slater.jl
def lda_slater_x(n, alpha=2 / 3, **kwargs):
    '''Slater exchange functional (spin-paired).

    Corresponds to the functional with the label LDA_C_PW and ID 1 in LibXC.
    Reference: Phys. Rev. 81, 385.

    Args:
        n (ndarray): Real-space electronic density.

    Keyword Args:
        alpha (float): Scaling factor.

    Returns:
        tuple[ndarray, ndarray]: Exchange energy density and potential.
    '''
    third = 1 / 3
    pi34 = (3 / (4 * np.pi))**third
    f = -9 / 8 * (3 / (2 * np.pi))**(2 / 3)
    rs = pi34 / n**third

    ex = f * alpha / rs
    vx = 4 / 3 * ex
    return ex, vx


# Adapted from
# https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/XC_funcs/XC_x_slater_spin.jl
def lda_slater_x_spin(n, zeta, alpha=2 / 3):
    '''Slater exchange functional (spin-polarized).

    Corresponds to the functional with the label LDA_C_PW and ID 1 in LibXC.
    Reference: Phys. Rev. 81, 385.

    Args:
        n (ndarray): Real-space electronic density.
        zeta (ndarray): Relative spin polarization.

    Keyword Args:
        alpha (float): Scaling factor.

    Returns:
        tuple[ndarray, ndarray]: Exchange energy density and potential.
    '''
    third = 1 / 3
    p43 = 4 / 3
    f = -9 / 8 * (3 / np.pi)**third

    rho13p = ((1 + zeta) * n)**third
    rho13m = ((1 - zeta) * n)**third

    exup = f * alpha * rho13p
    exdw = f * alpha * rho13m
    ex = 0.5 * ((1 + zeta) * exup + (1 - zeta) * exdw)

    vxup = p43 * f * alpha * rho13p
    vxdw = p43 * f * alpha * rho13m
    return ex, [vxup, vxdw]


# Adapted from https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/XC_funcs/XC_c_pw.jl
def lda_pw_c(n, **kwargs):
    '''Perdew-Wang parametrization of the correlation functional (spin-paired).

    Corresponds to the functional with the label LDA_C_PW and ID 12 in LibXC.
    Reference: Phys. Rev. B 45, 13244.

    Args:
        n (ndarray): Real-space electronic density.

    Returns:
        tuple[ndarray, ndarray]: PW correlation energy density and potential.
    '''
    third = 1 / 3
    pi34 = (3 / (4 * np.pi))**third
    rs = pi34 / n**third

    a = 0.031091
    a1 = 0.2137
    b1 = 7.5957
    b2 = 3.5876
    b3 = 1.6382
    b4 = 0.49294

    rs12 = np.sqrt(rs)
    rs32 = rs * rs12
    rs2 = rs**2

    om = 2 * a * (b1 * rs12 + b2 * rs + b3 * rs32 + b4 * rs2)
    olog = np.log(1 + 1 / om)
    ec = -2 * a * (1 + a1 * rs) * olog

    dom = 2 * a * (0.5 * b1 * rs12 + b2 * rs + 1.5 * b3 * rs32 + 2 * b4 * rs2)
    vc = -2 * a * (1 + 2 / 3 * a1 * rs) * olog - 2 / 3 * a * (1 + a1 * rs) * dom / (om * (om + 1))
    return ec, vc


# Adapted from https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/XC_funcs/XC_c_pw_spin.jl
def lda_pw_c_spin(n, zeta, **kwargs):
    '''Perdew-Wang parametrization of the correlation functional (spin-polarized).

    Corresponds to the functional with the label LDA_C_PW and ID 12 in LibXC.
    Reference: Phys. Rev. B 45, 13244.

    Args:
        n (ndarray): Real-space electronic density.
        zeta (ndarray): Relative spin polarization.

    Returns:
        tuple[ndarray, ndarray]: PW correlation energy density and potential.
    '''
    third = 1 / 3
    pi34 = (3 / (4 * np.pi))**third
    rs = pi34 / n**third

    fz0 = 1.709921

    zeta2 = zeta * zeta
    zeta3 = zeta2 * zeta
    zeta4 = zeta3 * zeta
    rs12 = np.sqrt(rs)
    rs32 = rs * rs12
    rs2 = rs**2

    def pw_fit(i):
        '''Calculate correlation energies by Perdew-Wang approximation interpolation.

        Args:
            i (int): Index to choose unpolarized (0), polarized (1), or antiferromagnetic (2) fit.

        Returns:
            tuple[ndarray, ndarray]: PW fit and the derivative.
        '''
        a = (0.031091, 0.015545, 0.016887)
        a1 = (0.2137, 0.20548, 0.11125)
        b1 = (7.5957, 14.1189, 10.357)
        b2 = (3.5876, 6.1977, 3.6231)
        b3 = (1.6382, 3.3662, 0.88026)
        b4 = (0.49294, 0.62517, 0.49671)

        om = 2 * a[i] * (b1[i] * rs12 + b2[i] * rs + b3[i] * rs32 + b4[i] * rs2)
        dom = 2 * a[i] * (0.5 * b1[i] * rs12 + b2[i] * rs + 1.5 * b3[i] * rs32 + 2 * b4[i] * rs2)
        olog = np.log(1 + 1 / om)

        fit = -2 * a[i] * (1 + a1[i] * rs) * olog
        dfit = -2 * a[i] * (1 + 2 / 3 * a1[i] * rs) * olog - \
               2 / 3 * a[i] * (1 + a1[i] * rs) * dom / (om * (om + 1))
        return fit, dfit

    ecU, vcU = pw_fit(0)  # Unpolarized
    ecP, vcP = pw_fit(1)  # Polarized
    ac, dac = pw_fit(2)  # Spin stiffness

    fz = ((1 + zeta)**(4 / 3) + (1 - zeta)**(4 / 3) - 2) / (2**(4 / 3) - 2)
    ec = ecU + ac * fz * (1 - zeta4) / fz0 + (ecP - ecU) * fz * zeta4

    dfz = ((1 + zeta)**third - (1 - zeta)**third) * 4 / (3 * (2**(4 / 3) - 2))
    vcup = vcU + dac * fz * (1 - zeta4) / fz0 + (vcP - vcU) * fz * zeta4 + \
           (ac / fz0 * (dfz * (1 - zeta4) - 4 * fz * zeta3) +
           (ecP - ecU) * (dfz * zeta4 + 4 * fz * zeta3)) * (1 - zeta)
    vcdw = vcU + dac * fz * (1 - zeta4) / fz0 + (vcP - vcU) * fz * zeta4 - \
           (ac / fz0 * (dfz * (1 - zeta4) - 4 * fz * zeta3) +
           (ecP - ecU) * (dfz * zeta4 + 4 * fz * zeta3)) * (1 + zeta)
    return ec, [vcup, vcdw]


# Adapted from https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/XC_funcs/XC_c_vwn.jl
def lda_vwn_c(n, **kwargs):
    '''Vosko, Wilk, and Nusair parametrization of the correlation functional (spin-paired).

    Corresponds to the functional with the label LDA_C_PW and ID 7 in LibXC.
    Reference: Phys. Rev. B 22, 3812.

    Args:
        n (ndarray): Real-space electronic density.

    Returns:
        tuple[ndarray, ndarray]: VWN correlation energy density and potential.
    '''
    third = 1 / 3
    pi34 = (3 / (4 * np.pi))**third
    rs = pi34 / n**third

    a = 0.0310907
    b = 3.72744
    c = 12.9352
    x0 = -0.10498

    q = np.sqrt(4 * c - b * b)
    f1 = 2 * b / q
    f2 = b * x0 / (x0 * x0 + b * x0 + c)
    f3 = 2 * (2 * x0 + b) / q
    rs12 = np.sqrt(rs)
    fx = rs + b * rs12 + c
    qx = np.arctan(q / (2 * rs12 + b))

    ec = a * (np.log(rs / fx) + f1 * qx - f2 * (np.log((rs12 - x0)**2 / fx) + f3 * qx))

    tx = 2 * rs12 + b
    tt = tx * tx + q * q
    vc = ec - rs12 * a / 6 * (2 / rs12 - tx / fx - 4 * b / tt -
         f2 * (2 / (rs12 - x0) - tx / fx - 4 * (2 * x0 + b) / tt))
    return ec, vc


# Adapted from https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/XC_funcs/XC_c_vwn_spin.jl
def lda_vwn_c_spin(n, zeta, **kwargs):
    '''Vosko, Wilk, and Nusair parametrization of the correlation functional (spin-polarized).

    Corresponds to the functional with the label LDA_C_PW and ID 7 in LibXC.
    Reference: Phys. Rev. B 22, 3812.

    Args:
        n (ndarray): Real-space electronic density.
        zeta (ndarray): Relative spin polarization.

    Returns:
        tuple[ndarray, ndarray]: VWN correlation energy density and potential.
    '''
    third = 1 / 3
    pi34 = (3 / (4 * np.pi))**third
    rs = pi34 / n**third

    cfz = 2**(4 / 3) - 2
    cfz1 = 1 / cfz
    cfz2 = 4 / 3 * cfz1
    iddfz0 = 9 / 8 * cfz
    sqrtrs = np.sqrt(rs)
    zeta3 = zeta**3
    zeta4 = zeta3 * zeta
    trup = 1 + zeta
    trdw = 1 - zeta
    trup13 = trup**third
    trdw13 = trdw**third
    fz = cfz1 * (trup13 * trup + trdw13 * trdw - 2)  # f(zeta)
    dfz = cfz2 * (trup13 - trdw13)  # df/dzeta

    def pade_fit(i):
        '''Calculate correlation energies by Pade approximation interpolation.

        Args:
            i (int): Index to choose paramagnetic (0), ferromagnetic (1), or spin stiffness (2) fit.

        Returns:
            tuple[ndarray, ndarray]: Pade fit and the derivative.
        '''
        a = (0.0310907, 0.01554535, -0.01688686394039)
        b = (3.72744, 7.06042, 1.13107)
        c = (12.9352, 18.0578, 13.0045)
        x0 = (-0.10498, -0.325, -0.0047584)
        Q = (6.15199081975908, 4.73092690956011, 7.12310891781812)
        tbQ = (1.21178334272806, 2.98479352354082, 0.31757762321188)
        bx0fx0 = (-0.03116760867894, -0.14460061018521, -0.00041403379428)

        xx0 = sqrtrs - x0[i]
        Qtxb = Q[i] / (2 * sqrtrs + b[i])
        atg = np.arctan(Qtxb)
        fx = rs + b[i] * sqrtrs + c[i]

        fit = a[i] * (np.log(rs / fx) + tbQ[i] * atg -
              bx0fx0[i] * (np.log(xx0 * xx0 / fx) + (tbQ[i] + 4 * x0[i] / Q[i]) * atg))

        txb = 2 * sqrtrs + b[i]
        txbfx = txb / fx
        itxbQ = 1 / (txb * txb + Q[i] * Q[i])

        dfit = fit - a[i] / 3 + a[i] * sqrtrs / 6 * (txbfx + 4 * b[i] * itxbQ +
               bx0fx0[i] * (2 / xx0 - txbfx - 4 * (b[i] + 2 * x0[i]) * itxbQ))
        return fit, dfit

    ecP, vcP = pade_fit(0)  # Paramagnetic fit
    ecF, vcF = pade_fit(1)  # Ferromagnetic fit
    ac, dac = pade_fit(2)  # Spin stiffness

    ac = ac * iddfz0
    De = ecF - ecP - ac  # e_c[F] - e_c[P] - alpha_c/(ddf/ddz(z=0))
    fzz4 = fz * zeta4
    ec = ecP + ac * fz + De * fzz4

    dac = dac * iddfz0
    dec1 = vcP + dac * fz + (vcF - vcP - dac) * fzz4  # e_c-(r_s/3)*(de_c/dr_s)
    dec2 = ac * dfz + De * (4 * zeta3 * fz + zeta4 * dfz)  # de_c/dzeta

    vcup = dec1 + (1 - zeta) * dec2
    vcdw = dec1 - (1 + zeta) * dec2
    return ec, [vcup, vcdw]



#!/usr/bin/env python3
'''Calculate different energy contributions.'''
import numpy as np
from numpy.linalg import inv
from scipy.special import erfc

from .dft import get_n_single, solve_poisson
from .xc import get_xc


class Energy:
    '''Energy class to save energy contributions in one place.'''
    __slots__ = ['Ekin', 'Eloc', 'Enonloc', 'Ecoul', 'Exc', 'Eewald', 'Esic']

    def __init__(self):
        self.Ekin = 0     # Kinetic energy
        self.Ecoul = 0    # Coulomb energy
        self.Exc = 0      # Exchange-correlation energy
        self.Eloc = 0     # Local energy
        self.Enonloc = 0  # Non-local energy
        self.Eewald = 0   # Ewald energy
        self.Esic = 0     # Self-interaction correction energy

    @property
    def Etot(self):
        '''Total energy is the sum of all energy contributions.'''
        return self.Ekin + self.Ecoul + self.Exc + self.Eloc + self.Enonloc + self.Eewald + \
               self.Esic

    def __repr__(self):
        '''Print the energies stored in the Energy object.'''
        out = ''
        for ie in self.__slots__:
            energy = eval('self.' + ie)
            if energy != 0:
                out = f'{out}{ie.ljust(8)}: {energy:+.9f} Eh\n'
        return f'{out}{"-" * 25}\nEtot    : {self.Etot:+.9f} Eh'


def get_E(scf):
    '''Calculate all energy contributions and update Atoms energies needed in one SCF step.

    Args:
        scf: SCF object.

    Returns:
        float: Total energy.
    '''
    scf.energies.Ekin = get_Ekin(scf.atoms, scf.Y)
    scf.energies.Ecoul = get_Ecoul(scf.atoms, scf.n, scf.phi)
    scf.energies.Exc = get_Exc(scf, scf.n, scf.exc)
    scf.energies.Eloc = get_Eloc(scf, scf.n)
    scf.energies.Enonloc = get_Enonloc(scf, scf.Y)
    return scf.energies.Etot


def get_Ekin(atoms, Y):
    '''Calculate the kinetic energy.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        Y (ndarray): Expansion coefficients of orthogonal wave functions in reciprocal space.

    Returns:
        float: Kinetic energy in Hartree.
    '''
    # Ekin = -0.5 Tr(F Wdag L(W))
    return np.real(-0.5 * np.trace(np.diag(atoms.f) @ (Y.conj().T @ atoms.L(Y))))


def get_Ecoul(atoms, n, phi=None):
    '''Calculate the Coulomb energy.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        n (ndarray): Real-space electronic density.

    Kwargs:
        phi (ndarray): Hartree ﬁeld.

    Returns:
        float: Coulomb energy in Hartree.
    '''
    if phi is None:
        phi = solve_poisson(atoms, n)
    # Ecoul = -(J(n))dag O(phi)
    return np.real(0.5 * n.conj().T @ atoms.Jdag(atoms.O(phi)))


def get_Exc(scf, n, exc=None):
    '''Calculate the exchange-correlation energy.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        scf: SCF object.
        n (ndarray): Real-space electronic density.

    Keyword Args:
        exc (ndarray): Exchange-correlation energy density.

    Returns:
        float: Exchange-correlation energy in Hartree.
    '''
    atoms = scf.atoms
    if exc is None:
        exc = get_xc(scf.xc, n)[0]
    # Exc = (J(n))dag O(J(exc))
    return np.real(n.conj().T @ atoms.Jdag(atoms.O(atoms.J(exc))))


def get_Eloc(scf, n):
    '''Calculate the local energy contribution.

    Args:
        scf: SCF object.
        n (ndarray): Real-space electronic density.

    Returns:
        float: Local energy contribution in Hartree.
    '''
    return np.real(scf.Vloc.conj().T @ n)


# Adapted from https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/calc_energies.jl
def get_Enonloc(scf, Y):
    '''Calculate the non-local GTH energy contribution.

    Reference: Phys. Rev. B 54, 1703.

    Args:
        scf: SCF object.
        Y (ndarray): Expansion coefficients of orthogonal wave functions in reciprocal space.

    Returns:
        float: Non-local GTH energy contribution in Hartree.
    '''
    atoms = scf.atoms
    Enonloc = 0
    if scf.NbetaNL > 0:  # Only calculate non-local potential if necessary
        Natoms = atoms.Natoms
        Nstates = atoms.Ns
        prj2beta = scf.prj2beta
        betaNL = scf.betaNL

        betaNL_psi = (Y.conj().T @ betaNL).conj()

        for ist in range(Nstates):
            enl = 0
            for ia in range(Natoms):
                psp = scf.GTH[atoms.atom[ia]]
                for l in range(psp['lmax']):
                    for m in range(-l, l + 1):
                        for iprj in range(psp['Nproj_l'][l]):
                            ibeta = prj2beta[iprj, ia, l, m + psp['lmax'] - 1] - 1
                            for jprj in range(psp['Nproj_l'][l]):
                                jbeta = prj2beta[jprj, ia, l, m + psp['lmax'] - 1] - 1
                                hij = psp['h'][l, iprj, jprj]
                                enl += hij * np.real(betaNL_psi[ist, ibeta].conj() *
                                       betaNL_psi[ist, jbeta])
            Enonloc += atoms.f[ist] * enl
    # We have to multiply with the cell volume, because of different orthogonalization methods
    return Enonloc * atoms.Omega


# Adapted from https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/calc_E_NN.jl
def get_Eewald(atoms, gcut=2, gamma=1e-8):
    '''Calculate the Ewald energy.

    Reference: J. Chem. Theory Comput. 10, 381.

    Args:
        atoms: Atoms object.

    Keyword Args:
        gcut (float): G-vector cut-off.
        gamma (float): Error tolerance

    Returns:
        float: Ewald energy in Hartree.
    '''
    # For a plane wave code we have multiple contributions for the Ewald energy
    # Namely, a sum from contributions from real-space, reciprocal space,
    # the self energy, (the dipole term [neglected]), and an additional electroneutrality term
    Natoms = atoms.Natoms
    X = atoms.X
    Z = atoms.Z
    Omega = atoms.Omega
    R = atoms.R

    t1, t2, t3 = R
    t1m = np.sqrt(t1 @ t1)
    t2m = np.sqrt(t2 @ t2)
    t3m = np.sqrt(t3 @ t3)

    g1, g2, g3 = 2 * np.pi * inv(R.conj().T)
    g1m = np.sqrt(g1 @ g1)
    g2m = np.sqrt(g2 @ g2)
    g3m = np.sqrt(g3 @ g3)

    gexp = -np.log(gamma)
    nu = 0.5 * np.sqrt(gcut**2 / gexp)

    x = np.sum(Z**2)
    totalcharge = np.sum(Z)

    # Start by calculating the self-energy
    Eewald = -nu * x / np.sqrt(np.pi)
    # Add the electroneutrality term
    Eewald += -np.pi * totalcharge**2 / (2 * Omega * nu**2)

    tmax = np.sqrt(0.5 * gexp) / nu
    mmm1 = np.rint(tmax / t1m + 1.5)
    mmm2 = np.rint(tmax / t2m + 1.5)
    mmm3 = np.rint(tmax / t3m + 1.5)

    dX = np.empty(3)
    T = np.empty(3)
    for ia in range(Natoms):
        for ja in range(Natoms):
            dX = X[ia] - X[ja]
            ZiZj = Z[ia] * Z[ja]
            for i in np.arange(-mmm1, mmm1 + 1):
                for j in np.arange(-mmm2, mmm2 + 1):
                    for k in np.arange(-mmm3, mmm3 + 1):
                        if (ia != ja) or ((abs(i) + abs(j) + abs(k)) != 0):
                            T = i * t1 + j * t2 + k * t3
                            rmag = np.sqrt(np.sum((dX - T)**2))
                            # Add the real-space contribution
                            Eewald += 0.5 * ZiZj * erfc(rmag * nu) / rmag

    mmm1 = np.rint(gcut / g1m + 1.5)
    mmm2 = np.rint(gcut / g2m + 1.5)
    mmm3 = np.rint(gcut / g3m + 1.5)

    G = np.empty(3)
    for ia in range(Natoms):
        for ja in range(Natoms):
            dX = X[ia] - X[ja]
            ZiZj = Z[ia] * Z[ja]
            for i in np.arange(-mmm1, mmm1 + 1):
                for j in np.arange(-mmm2, mmm2 + 1):
                    for k in np.arange(-mmm3, mmm3 + 1):
                        if (abs(i) + abs(j) + abs(k)) != 0:
                            G = i * g1 + j * g2 + k * g3
                            GX = np.sum(G * dX)
                            G2 = np.sum(G**2)
                            # Add the reciprocal space contribution
                            x = 2 * np.pi / Omega * np.exp(-0.25 * G2 / nu**2) / G2
                            Eewald += x * ZiZj * np.cos(GX)
    return Eewald


def get_Esic(scf, Y, n_single=None):
    '''Calculate the Perdew-Zunger self-interaction energy.

    Reference: Phys. Rev. B 23, 5048.

    Args:
        scf: SCF object.
        Y (ndarray): Expansion coefficients of orthogonal wave functions in reciprocal space.

    Keyword Args:
        n_single (ndarray): Single-electron densities.

    Returns:
        float: PZ self-interaction energy.
    '''
    atoms = scf.atoms
    # E_PZ-SIC = \sum_i Ecoul[n_i] + Exc[n_i, 0]
    if n_single is None:
        n_single = get_n_single(atoms, Y)

    Esic = 0
    for i in range(atoms.Ns):
        # Normalize single-particle densities to 1
        ni = n_single[:, i] / atoms.f[i]
        coul = get_Ecoul(atoms, ni)
        # The exchange part for a SIC correction has to be spin polarized
        xc = get_Exc(scf, ni, spinpol=True)
        # SIC energy is scaled by the occupation number
        Esic += (coul + xc) * atoms.f[i]
    scf.energies.Esic = Esic
    return Esic



#!/usr/bin/env python3
'''Interface to use LibXC functionals.

For a list of available functionals, see: https://www.tddft.org/programs/libxc/functionals/
'''
try:
    from pylibxc import LibXCFunctional
except ImportError:
    print('ERROR: Necessary addon dependencies not found. To use this module,\n'
          '       install the package with addons, e.g., with "pip install eminus[addons]"')

from ..logger import log


def libxc_functional(xc, n, spinpol):
    '''Handle LibXC exchange-correlation functionals.

    Only LDA functionals should be used.

    Reference: SoftwareX 7, 1.

    Args:
        xc (str | int): Exchange or correlation identifier.
        n (ndarray): Real-space electronic density.
        spinpol (bool): Choose if a spin-polarized exchange-correlation functional will be used.

    Returns:
        tuple[ndarray, ndarray]: Exchange-correlation energy density and potential.
    '''
    # FIXME: Compare LibXC and internal polarized functionals
    if spinpol:
        log.warning('The LibXC routine will still use an unpolarized functional.')
    spin = 'unpolarized'

    inp = {'rho': n}
    # LibXC functionals have one integer and one string identifier
    try:
        func = LibXCFunctional(int(xc), spin)
    except ValueError:
        func = LibXCFunctional(xc, spin)
    out = func.compute(inp)

    exc = out['zk'].ravel()
    vxc = out['vrho'].ravel()
    return exc, vxc



#!/usr/bin/env python3
'''Logger initialization and configuration.'''
import logging
import sys


class CustomLogger(logging.Logger):
    '''Custom logger for the usage outside of classes.

    This is just a basic logger, but with an added verbose property.

    Args:
        name (str): Logger name.
    '''
    def __init__(self, name):
        super(CustomLogger, self).__init__(name)

    @property
    def verbose(self):
        '''Verbosity level.'''
        return self._verbose

    @verbose.setter
    def verbose(self, level):
        '''Verbosity setter to sync the logger with the property.'''
        self._verbose = get_level(level)
        self.setLevel(self._verbose)


class CustomFormatter(logging.Formatter):
    '''Custom logger formatter.

    Inherited from logging.Formatter.
    '''
    def format(self, record):
        '''Use different formatting for different logging levels.

        Args:
            record: LogRecord object.
        '''
        if record.levelno >= logging.WARNING:
            # Print the level name for errors and warning
            self._style._fmt = '%(levelname)s: %(msg)s'
        else:
            # But not for infos and debug messages
            self._style._fmt = '%(msg)s'
        return super().format(record)


def create_logger(object):
    '''Create a logger unique to an object.

    Args:
        object: Instance of a class.
    '''
    # Use ID of object to create a unique logger
    # Without this setting the verbosity in one instance would affect other instances
    return logging.getLogger(str(id(object)))


def get_level(verbose):
    '''Validate logging levels.

    Args:
        verbose (int | str): Level of output (case insensitive).
    '''
    log_levels = {
        0: 'CRITICAL',
        1: 'ERROR',
        2: 'WARNING',
        3: 'INFO',
        4: 'DEBUG'
    }
    if isinstance(verbose, int):
        level = log_levels.get(verbose, 'DEBUG')
    else:
        level = verbose.upper()
    return level


def name(newname):
    '''Add a name to functions without evaluating them for better logging.

    Args:
        newname (str): Function name.

    Returns:
        Callable: Decorator.
    '''
    def decorator(f):
        '''Decorator.

        Args:
            f (Callable): Function to name.

        Returns:
            Callable: Function.
        '''
        f.__name__ = newname
        return f
    return decorator


# The following code is not guarded by a function because it has to be run once the logger is called
# to set up the basic logger configuration

# Create a base logger that can be used outside of classes
logging.setLoggerClass(CustomLogger)
log = logging.getLogger('eminus')
log.verbose = 'WARNING'

# Basic logger setup
formatter = CustomFormatter()
handler = logging.StreamHandler(sys.stdout)
handler.setFormatter(formatter)
logging.root.addHandler(handler)



#!/usr/bin/env python3
'''Workflow functions that combine functions to generate orbitals.'''
from .dft import get_psi
from .filehandler import write_cube
from .localizer import get_FLO, get_FO
from .logger import log


def KSO(scf, write_cubes=False, **kwargs):
    '''Generate Kohn-Sham orbitals and optionally save them as cube files.

    Args:
        scf: SCF object.

    Keyword Args:
        write_cubes (bool): Write orbitals to cube files.

    Returns:
        ndarray: Real-space Kohn-Sham orbitals.
    '''
    atoms = scf.atoms

    # Calculate eigenfunctions and transform to real-space
    KSO = atoms.I(get_psi(scf, scf.W))
    if write_cubes:
        cube_writer(atoms, KSO, 'KSO')
    return KSO


def FO(scf, write_cubes=False, fods=None):
    '''Generate Fermi orbitals and optionally save them as cube files.

    Reference: J. Chem. Phys. 153, 084104.

    Args:
        scf: SCF object.

    Keyword Args:
        write_cubes (bool): Write orbitals to cube files.
        fods (ndarray): Fermi-orbital descriptors.

    Returns:
        ndarray: Real-space Fermi orbitals.
    '''
    # Lazy import addons
    from .addons.fods import get_fods, remove_core_fods
    atoms = scf.atoms

    # Calculate eigenfunctions
    KSO = get_psi(scf, scf.W)
    if fods is None:
        fods = get_fods(atoms)
    fods = remove_core_fods(atoms, fods)

    # The FO functions needs orbitals in reciprocal space as input
    FO = get_FO(atoms, KSO, fods)
    if write_cubes:
        cube_writer(atoms, FO, 'FO')
    return FO


def FLO(scf, write_cubes=False, fods=None):
    '''Generate Fermi-Löwdin orbitals and optionally save them as cube files.

    Reference: J. Chem. Phys. 153, 084104.

    Args:
        scf: SCF object.

    Keyword Args:
        write_cubes (bool): Write orbitals to cube files.
        fods (ndarray): Fermi-orbital descriptors.

    Returns:
        ndarray: Real-space Fermi-Löwdin orbitals.
    '''
    # Lazy import addons
    from .addons.fods import get_fods, remove_core_fods
    atoms = scf.atoms

    # Calculate eigenfunctions
    KSO = get_psi(scf, scf.W)
    if fods is None:
        fods = get_fods(atoms)
    fods = remove_core_fods(atoms, fods)

    # The FLO functions needs orbitals in reciprocal space as input
    FLO = get_FLO(atoms, KSO, fods)
    if write_cubes:
        cube_writer(atoms, FLO, 'FLO')
    return FLO


def cube_writer(atoms, orbitals, type):
    '''Simple cube file writer function.

    Args:
        atoms: Atoms object.
        orbitals (ndarray): Real-space orbitals.
        type (str): Orbital type for the cube file names.
    '''
    # Create the system name
    name = ''
    for ia in set(atoms.atom):
        # Skip the number of atoms if it is equal to one
        na = atoms.atom.count(ia)
        name += f'{ia}{na if na > 1 else ""}'

    for i in range(atoms.Ns):
        log.info(f'Write {name}_{type}_{i}.cube...')
        write_cube(atoms, orbitals[:, i], f'{name}_{type}_{i}.cube')
    return



#!/usr/bin/env python3
'''eminus - A plane wave density functional theory code.

Minimal usage example to do a DFT calculation for helium::

   from eminus import Atoms, SCF
   atoms = Atoms('He', [0, 0, 0])
   SCF(atoms)
'''
from .atoms import Atoms
from .dft import get_epsilon, get_psi
from .filehandler import load, read_cube, read_xyz, save, write_cube, write_xyz
from .logger import log
from .scf import SCF
from .version import __version__, info

__all__ = ['Atoms', 'get_epsilon', 'get_psi', 'info', 'load', 'log', 'read_cube', 'read_xyz',
           'save', 'SCF', 'write_cube', 'write_xyz', '__version__']



#!/usr/bin/env python3
'''Minimization algorithms.'''
import logging

import numpy as np

from .dft import get_grad, get_n_total, orth, solve_poisson
from .energies import get_E
from .logger import name
from .utils import dotprod
from .xc import get_xc


def scf_step(scf):
    '''Perform one SCF step for a DFT calculation.

    Calculating intermediate results speeds up the energy and gradient calculation.

    Args:
        scf: SCF object.

    Returns:
        float: Total energy.
    '''
    scf.Y = orth(scf.atoms, scf.W)
    scf.n = get_n_total(scf.atoms, scf.Y)
    scf.phi = solve_poisson(scf.atoms, scf.n)
    scf.exc, scf.vxc = get_xc(scf.xc, scf.n)
    return get_E(scf)


def check_energies(scf, Elist, linmin='', cg=''):
    '''Check the energies for every SCF cycle and handle the output.

    Args:
        scf: SCF object.
        Elist (list): Total energies per SCF step.

    Keyword Args:
        linmin (float): Cosine between previous search direction and current gradient.
        cg (float): Conjugate-gradient orthogonality.

    Returns:
        bool: Convergence condition.
    '''
    iteration = len(Elist)

    # Output handling
    if not isinstance(linmin, str):
        linmin = f' \tlinmin-test: {linmin:+.7f}'
    if not isinstance(cg, str):
        cg = f' \tcg-test: {cg:+.7f}'
    if scf.log.level <= logging.DEBUG:
        scf.log.debug(f'Iteration: {iteration}  \tEtot: {scf.energies.Etot:+.7f}{linmin}{cg}')
    else:
        scf.log.info(f'Iteration: {iteration}  \tEtot: {scf.energies.Etot:+.7f}')

    # Check for convergence
    if iteration > 1:
        if abs(Elist[-2] - Elist[-1]) < scf.etol:
            return True
        elif Elist[-1] > Elist[-2]:
            scf.log.warning('Total energy is not decreasing.')
    return False


@name('steepest descent minimization')
def sd(scf, Nit, cost=scf_step, grad=get_grad, condition=check_energies, betat=3e-5):
    '''Steepest descent minimization algorithm.

    Args:
        scf: SCF object.
        Nit (int): Maximum number of SCF steps.

    Keyword Args:
        cost (Callable): Function that will run every SCF step.
        grad (Callable): Function that calculates the respective gradient.
        condition (Callable): Function to check and log the convergence condition.
        betat (float): SCF step size.

    Returns:
        list: Total energies per SCF cycle.
    '''
    costs = []

    for _ in range(Nit):
        c = cost(scf)
        costs.append(c)
        if condition(scf, costs):
            break
        g = grad(scf, scf.W, scf.Y, scf.n, scf.phi, scf.vxc)
        scf.W = scf.W - betat * g
    return costs


@name('line minimization')
def lm(scf, Nit, cost=scf_step, grad=get_grad, condition=check_energies, betat=3e-5):
    '''Line minimization algorithm.

    Args:
        scf: SCF object.
        Nit (int): Maximum number of SCF steps.

    Keyword Args:
        cost (Callable): Function that will run every SCF step.
        grad (Callable): Function that calculates the respective gradient.
        condition (Callable): Function to check and log the convergence condition.
        betat (float): SCF step size.

    Returns:
        list: Total energies per SCF cycle.
    '''
    costs = []
    linmin = ''

    # Do the first step without the linmin test
    g = grad(scf, scf.W)
    d = -g
    gt = grad(scf, scf.W + betat * d)
    beta = betat * dotprod(g, d) / dotprod(g - gt, d)

    scf.W = scf.W + beta * d
    c = cost(scf)
    costs.append(c)
    condition(scf, costs)

    for _ in range(1, Nit):
        g = grad(scf, scf.W, scf.Y, scf.n, scf.phi, scf.vxc)
        if scf.log.level <= logging.DEBUG:
            linmin = dotprod(g, d) / np.sqrt(dotprod(g, g) * dotprod(d, d))
        d = -g
        gt = grad(scf, scf.W + betat * d)
        beta = betat * dotprod(g, d) / dotprod(g - gt, d)

        scf.W = scf.W + beta * d
        c = cost(scf)
        costs.append(c)
        if condition(scf, costs, linmin):
            break
    return costs


@name('preconditioned line minimization')
def pclm(scf, Nit, cost=scf_step, grad=get_grad, condition=check_energies, betat=3e-5):
    '''Preconditioned line minimization algorithm.

    Args:
        scf: SCF object.
        Nit (int): Maximum number of SCF steps.

    Keyword Args:
        cost (Callable): Function that will run every SCF step.
        grad (Callable): Function that calculates the respective gradient.
        condition (Callable): Function to check and log the convergence condition.
        betat (float): SCF step size.

    Returns:
        list: Total energies per SCF cycle.
    '''
    atoms = scf.atoms
    costs = []
    linmin = ''

    # Do the first step without the linmin test
    g = grad(scf, scf.W)
    d = -atoms.K(g)
    gt = grad(scf, scf.W + betat * d)
    beta = betat * dotprod(g, d) / dotprod(g - gt, d)

    scf.W = scf.W + beta * d
    c = cost(scf)
    costs.append(c)
    condition(scf, costs)

    for _ in range(1, Nit):
        g = grad(scf, scf.W, scf.Y, scf.n, scf.phi, scf.vxc)
        if scf.log.level <= logging.DEBUG:
            linmin = dotprod(g, d) / np.sqrt(dotprod(g, g) * dotprod(d, d))
        d = -atoms.K(g)
        gt = grad(scf, scf.W + betat * d)
        beta = betat * dotprod(g, d) / dotprod(g - gt, d)

        scf.W = scf.W + beta * d
        c = cost(scf)
        costs.append(c)
        if condition(scf, costs, linmin):
            break
    return costs


@name('conjugate-gradient minimization')
def cg(scf, Nit, cost=scf_step, grad=get_grad, condition=check_energies, betat=3e-5):
    '''Conjugate-gradient minimization algorithm.

    Args:
        scf: SCF object.
        Nit (int): Maximum number of SCF steps.

    Keyword Args:
        cost (Callable): Function that will run every SCF step.
        grad (Callable): Function that calculates the respective gradient.
        condition (Callable): Function to check and log the convergence condition.
        betat (float): SCF step size.

    Returns:
        list: Total energies per SCF cycle.
    '''
    costs = []
    linmin = ''
    cg = ''

    # Do the first step without the linmin and cg test
    g = grad(scf, scf.W)
    d = -g
    gt = grad(scf, scf.W + betat * d)
    beta = betat * dotprod(g, d) / dotprod(g - gt, d)
    d_old = d
    g_old = g

    scf.W = scf.W + beta * d
    c = cost(scf)
    costs.append(c)
    condition(scf, costs)

    for _ in range(1, Nit):
        g = grad(scf, scf.W, scf.Y, scf.n, scf.phi, scf.vxc)
        if scf.log.level <= logging.DEBUG:
            linmin = dotprod(g, d_old) / np.sqrt(dotprod(g, g) * dotprod(d_old, d_old))
            cg = dotprod(g, g_old) / np.sqrt(dotprod(g, g) *
                 dotprod(g_old, g_old))
        if scf.cgform == 1:  # Fletcher-Reeves
            beta = dotprod(g, g) / dotprod(g_old, g_old)
        elif scf.cgform == 2:  # Polak-Ribiere
            beta = dotprod(g - g_old, g) / dotprod(g_old, g_old)
        elif scf.cgform == 3:  # Hestenes-Stiefel
            beta = dotprod(g - g_old, g) / dotprod(g - g_old, d_old)
        d = -g + beta * d_old
        gt = grad(scf, scf.W + betat * d)
        beta = betat * dotprod(g, d) / dotprod(g - gt, d)
        d_old = d
        g_old = g

        scf.W = scf.W + beta * d
        c = cost(scf)
        costs.append(c)
        if condition(scf, costs, linmin, cg):
            break
    return costs


@name('preconditioned conjugate-gradient minimization')
def pccg(scf, Nit, cost=scf_step, grad=get_grad, condition=check_energies, betat=3e-5):
    '''Preconditioned conjugate-gradient minimization algorithm.

    Args:
        scf: SCF object.
        Nit (int): Maximum number of SCF steps.

    Keyword Args:
        cost (Callable): Function that will run every SCF step.
        grad (Callable): Function that calculates the respective gradient.
        condition (Callable): Function to check and log the convergence condition.
        betat (float): SCF step size.

    Returns:
        list: Total energies per SCF cycle.
    '''
    atoms = scf.atoms
    costs = []
    linmin = ''
    cg = ''

    # Do the first step without the linmin and cg test
    g = grad(scf, scf.W)
    d = -atoms.K(g)
    gt = grad(scf, scf.W + betat * d)
    beta = betat * dotprod(g, d) / dotprod(g - gt, d)
    d_old = d
    g_old = g

    scf.W = scf.W + beta * d
    c = cost(scf)
    costs.append(c)
    condition(scf, costs)

    for _ in range(1, Nit):
        g = grad(scf, scf.W, scf.Y, scf.n, scf.phi, scf.vxc)
        if scf.log.level <= logging.DEBUG:
            linmin = dotprod(g, d_old) / np.sqrt(dotprod(g, g) * dotprod(d_old, d_old))
            cg = dotprod(g, atoms.K(g_old)) / np.sqrt(dotprod(g, atoms.K(g)) *
                 dotprod(g_old, atoms.K(g_old)))
        if scf.cgform == 1:  # Fletcher-Reeves
            beta = dotprod(g, atoms.K(g)) / dotprod(g_old, atoms.K(g_old))
        elif scf.cgform == 2:  # Polak-Ribiere
            beta = dotprod(g - g_old, atoms.K(g)) / dotprod(g_old, atoms.K(g_old))
        elif scf.cgform == 3:  # Hestenes-Stiefel
            beta = dotprod(g - g_old, atoms.K(g)) / dotprod(g - g_old, d_old)
        d = -atoms.K(g) + beta * d_old
        gt = grad(scf, scf.W + betat * d)
        beta = betat * dotprod(g, d) / dotprod(g - gt, d)
        d_old = d
        g_old = g

        scf.W = scf.W + beta * d
        c = cost(scf)
        costs.append(c)
        if condition(scf, costs, linmin, cg):
            break
    return costs



#!/usr/bin/env python3
'''Import and export functionalities.'''
import glob
import os
import pickle
import textwrap
import time

import numpy as np

from . import __path__
from .data import number2symbol, symbol2number
from .logger import log
from .units import ang2bohr, bohr2ang
from .version import __version__


def read_xyz(filename):
    '''Load atom species and positions from xyz files.

    File format definition: https://openbabel.org/wiki/XYZ_%28format%29

    Args:
        filename (str): xyz input file path/name.

    Returns:
        tuple[list, ndarray]: Atom species and positions.
    '''
    if not filename.endswith('.xyz'):
        filename = f'{filename}.xyz'

    with open(filename, 'r') as fh:
        lines = fh.readlines()

        # The first line contains the number of atoms
        Natoms = int(lines[0].strip())

        # The second line can contain a comment, print it if available
        comment = lines[1].strip()
        log.info(f'XYZ file comment: "{comment}"')

        atom = []
        X = []
        # Following lines contain atom positions with the format: Atom x-pos y-pos z-pos
        for line in lines[2:2 + Natoms]:
            line_split = line.strip().split()
            atom.append(line_split[0])
            X.append(np.float_(line_split[1:4]))

    # xyz files are in Angstrom, so convert to Bohr
    X = ang2bohr(np.asarray(X))
    return atom, X


def write_xyz(object, filename, extra=None):
    '''Generate xyz files from atoms objects.

    File format definition: https://openbabel.org/wiki/XYZ_%28format%29

    Args:
        object: Atoms or SCF object.
        filename (str): xyz output file path/name.

    Keyword Args:
        extra (ndarray): Extra coordinates to write.
    '''
    try:
        atoms = object.atoms
    except AttributeError:
        atoms = object
    atom = atoms.atom
    Natoms = atoms.Natoms
    X = atoms.X

    if not filename.endswith('.xyz'):
        filename = f'{filename}.xyz'

    # Convert the coordinates from atomic units to Angstrom
    X = bohr2ang(X)
    if extra is not None:
        extra = bohr2ang(extra)

    with open(filename, 'w') as fp:
        # The first line contains the number of atoms.
        # If we add extra coordinates, add them to the count.
        if extra is None:
            fp.write(f'{Natoms}\n')
        else:
            fp.write(f'{Natoms + len(extra)}\n')
        # The second line can contains a comment.
        # Print information about the file and program, and the file creation time.
        fp.write(f'File generated with eminus {__version__} on {time.ctime()}\n')
        for ia in range(Natoms):
            fp.write(f'{atom[ia]}  {X[ia][0]:.5f}  {X[ia][1]:.5f}  {X[ia][2]:.5f}\n')
        # Add extra coordinates if desired. The atom symbol will default to X (no atom type).
        if extra is not None:
            for ie in extra:
                fp.write(f'X  {ie[0]:.5f}  {ie[1]:.5f}  {ie[2]:.5f}\n')
    return


def read_cube(filename):
    '''Load atom and cell data from cube files.

    There is no standard for cube files. The following format has been used.
    File format definition: https://h5cube-spec.readthedocs.io/en/latest/cubeformat.html

    Args:
        filename (str): cube input file path/name.

    Returns:
        tuple[list, ndarray, float, ndarray, int]: Species, positions, charges, cell size, sampling.
    '''
    if not filename.endswith('.cube'):
        filename = f'{filename}.cube'

    # Atomic units and a cuboidal unit cell that starts at (0,0,0) are assumed.
    with open(filename, 'r') as fh:
        lines = fh.readlines()

        # The first and second line can contain comments, print them if available
        comment = f'{lines[0].strip()}\n{lines[1].strip()}'
        log.info(f'XYZ file comment: "{comment}"')

        # Line 4 to 6 contain the sampling per axis, and the unit cell basis vectors with length a/s
        # A cuboidal unit cell is assumed, so only use the diagonal entries
        s = np.empty(3, dtype=int)
        a = np.empty(3)
        for i, line in enumerate(lines[3:6]):
            line_split = line.strip().split()
            s[i] = line_split[0]
            a[i] = s[i] * np.float_(line_split[i + 1])

        atom = []
        X = []
        Z = []
        # Following lines contain atom positions with the format: atom-id charge x-pos y-pos z-pos
        for line in lines[6:]:
            line_split = line.strip().split()
            # If the first value is not a (positive) integer, we have reached the field data
            if not line_split[0].isdigit():
                break
            atom.append(number2symbol[int(line_split[0])])
            Z.append(line_split[1])
            X.append(np.float_(line_split[2:5]))

    X = np.asarray(X)
    return atom, X, Z, a, s


def write_cube(object, field, filename, extra=None):
    '''Generate cube files from given field data.

    There is no standard for cube files. The following format has been used to work with VESTA.
    File format definition: https://h5cube-spec.readthedocs.io/en/latest/cubeformat.html

    Args:
        object: Atoms or SCF object.
        field (ndarray): Real-space field data.
        filename (str): xyz output file path/name.

    Keyword Args:
        extra (ndarray): Extra coordinates to write.
    '''
    try:
        atoms = object.atoms
    except AttributeError:
        atoms = object
    # Atomic units are assumed, so there is no need for conversion.
    atom = atoms.atom
    Natoms = atoms.Natoms
    a = atoms.a
    r = atoms.r
    s = atoms.s
    X = atoms.X
    Z = atoms.Z

    if not filename.endswith('.cube'):
        filename = f'{filename}.cube'

    # Our field data has been created in a different order than needed for cube files
    # (triple loop over z,y,x instead of x,y,z), so rearrange it with some index magic.
    idx = []
    for Nx in range(s[0]):
        for Ny in range(s[1]):
            for Nz in range(s[2]):
                idx.append(Nx + Ny * s[0] + Nz * s[0] * s[1])
    idx = np.asarray(idx)

    # Make sure we have real valued data in the correct order
    field = np.real(field[idx])

    with open(filename, 'w') as fp:
        # The first two lines have to be a comment.
        # Print information about the file and program, and the file creation time.
        fp.write(f'File generated with eminus {__version__} on {time.ctime()}\n\n')
        # Number of atoms (int), and origin of the coordinate system (float)
        # The origin is normally at 0,0,0 but we could move our box, so take the minimum
        if extra is None:
            fp.write(f'{Natoms}  ')
        else:
            fp.write(f'{Natoms + len(extra)}  ')
        fp.write(f'{min(r[:, 0]):.5f}  {min(r[:, 1]):.5f}  {min(r[:, 2]):.5f}\n')
        # Number of points per axis (int), and vector defining the axis (float)
        # We only have a cuboidal box, so each vector only has one non-zero component
        fp.write(f'{s[0]}  {a[0] / s[0]:.5f}  0.0  0.0\n')
        fp.write(f'{s[1]}  0.0  {a[1] / s[1]:.5f}  0.0\n')
        fp.write(f'{s[2]}  0.0  0.0  {a[2] / s[2]:.5f}\n')
        # Atomic number (int), atomic charge (float), and atom position (floats) for every atom
        for ia in range(Natoms):
            fp.write(f'{symbol2number[atom[ia]]}  {Z[ia]:.5f}  ')
            fp.write(f'{X[ia][0]:.5f}  {X[ia][1]:.5f}  {X[ia][2]:.5f}\n')
        if extra is not None:
            for ie in extra:
                fp.write('0  0.00000  ')
                fp.write(f'{ie[0]:.5f}  {ie[1]:.5f}  {ie[2]:.5f}\n')
        # Field data (float) with scientific formatting
        # We have s[0]*s[1] chunks values with a length of s[2]
        for i in range(s[0] * s[1]):
            # Print every round of values, so we can add empty lines between them
            data_str = '%+1.5e  ' * s[2] % tuple(field[i * s[2]:(i + 1) * s[2]])
            # Print a maximum of 6 values per row
            # Max width for this formatting is 90, since 6*len('+1.00000e-000  ')=90
            fp.write(f'{textwrap.fill(data_str, width=90)}\n\n')
    return


def save(object, filename):
    '''Save objects in a pickle file.

    This function is for personal use only. Never load a file you haven't saved yourself!

    Args:
        object: Class object.
        filename (str): xyz input file path/name.
    '''
    if not filename.endswith(('.pickle', '.pkl')):
        filename = f'{filename}.pickle'

    with open(filename, 'wb') as fp:
        pickle.dump(object, fp, pickle.HIGHEST_PROTOCOL)
    return


def load(filename):
    '''Load objects from a pickle file.

    This function is for personal use only. Never load a file you haven't saved yourself!

    Args:
        filename (str): xyz input file path/name.

    Returns:
        Class object.
    '''
    if not filename.endswith(('.pickle', '.pkl')):
        filename = f'{filename}.pickle'

    with open(filename, 'rb') as fh:
        return pickle.load(fh)


def create_pdb(atom, X, a=None):
    '''Convert atom symbols and positions to the pdb format.

    File format definitions:
        CRYST1: https://www.wwpdb.org/documentation/file-format-content/format33/sect8.html#CRYST1

        ATOM: https://www.wwpdb.org/documentation/file-format-content/format33/sect9.html#ATOM

    Args:
        atom (list): Atom symbols.
        X (ndarray): Atom positions.

    Keyword Args:
        a (float): Cell size.

    Returns:
        str: pdb file format.
    '''
    # Convert Bohr to Angstrom
    X = bohr2ang(X)
    if a is not None:
        a = bohr2ang(a)

    # pdb files have specific numbers of characters for every data with changing justification
    # Write everything explicitly down to not loose track of line lengths
    pdb = ''
    # Create data for a cuboidal cell
    if a is not None:
        pdb += 'CRYST1'                 # 1-6 "CRYST1"
        pdb += f'{a[0]:9.3f}'.rjust(9)  # 7-15 a
        pdb += f'{a[1]:9.3f}'.rjust(9)  # 16-24 b
        pdb += f'{a[2]:9.3f}'.rjust(9)  # 25-33 c
        pdb += f'{90:7.2f}'.rjust(7)    # 34-40 alpha
        pdb += f'{90:7.2f}'.rjust(7)    # 41-47 beta
        pdb += f'{90:7.2f}'.rjust(7)    # 48-54 gamma
        pdb += ' '
        pdb += 'P 1'.ljust(11)          # 56-66 Space group
        # pdb += '1'.rjust(4)           # 67-70 Z value

    # Create molecule data
    pdb += '\nMODEL 1'
    for ia in range(len(atom)):
        pdb += '\nATOM  '                   # 1-6 "ATOM"
        pdb += f'{ia + 1}'.rjust(5)         # 7-11 Atom serial number
        pdb += ' '
        pdb += f'{atom[ia]}'.rjust(4)       # 13-16 Atom name
        pdb += ' '                          # 17 Alternate location indicator
        pdb += 'MOL'                        # 18-20 Residue name
        pdb += ' '
        pdb += ' '                          # 22 Chain identifier
        pdb += '1'.rjust(4)                 # 23-26 Residue sequence number
        pdb += ' '                          # 27 Code for insertions of residues
        pdb += '   '
        pdb += f'{X[ia][0]:8.3f}'.rjust(8)  # 31-38 X orthogonal coordinate
        pdb += f'{X[ia][1]:8.3f}'.rjust(8)  # 39-46 Y orthogonal coordinate
        pdb += f'{X[ia][2]:8.3f}'.rjust(8)  # 47-54 Z orthogonal coordinate
        pdb += f'{1:6.2f}'.rjust(6)         # 55-60 Occupancy
        pdb += f'{0:6.2f}'.rjust(6)         # 61-66 Temperature factor
        pdb += '          '
        pdb += f'{atom[ia]}'.rjust(2)       # 77-78 Element symbol
        # pdb += '  '                       # 79-80 Charge
    return f'{pdb}\nENDMDL'


def read_gth(atom, charge=None, psp_path=None):
    '''Read GTH files for a given atom.

    Args:
        atom (str): Atom name.

    Keyword Args:
        charge (int): Valence charge.
        psp_path (str): Path to GTH pseudopotential files. Defaults to installation_path/pade_gth/.

    Returns:
        dict: GTH parameters.
    '''
    if psp_path is None:
        psp_path = f'{__path__[0]}/pade_gth/'

    if charge is not None:
        f_psp = f'{psp_path}{atom}-q{charge}.gth'
    else:
        files = glob.glob(f'{psp_path}{atom}-q*')
        files.sort()
        try:
            f_psp = files[0]
        except IndexError:
            log.exception(f'There is no GTH pseudopotential in {psp_path} for "{atom}"')
            raise
        if len(files) > 1:
            log.info(f'Multiple pseudopotentials found for "{atom}". '
                     f'Continue with "{os.path.basename(f_psp)}".')

    psp = {}
    cloc = np.zeros(4)
    rp = np.zeros(4)
    Nproj_l = np.zeros(4, dtype=int)
    h = np.zeros([4, 3, 3])
    try:
        with open(f_psp, 'r') as fh:
            # Skip the first line, since we don't need the atom symbol here. If needed, use
            # psp['atom'] = fh.readline().split()[0]  # Atom symbol
            fh.readline()
            N_all = fh.readline().split()
            N_s, N_p, N_d, N_f = int(N_all[0]), int(N_all[1]), int(N_all[2]), int(N_all[3])
            psp['Zion'] = N_s + N_p + N_d + N_f  # Ionic charge
            # Skip the number of local coefficients, since we don't need it. If needed, use
            # rloc, n_c_local = fh.readline().split()
            # psp['n_c_local'] = int(n_c_local)  # Number of local coefficients
            rloc = fh.readline().split()[0]
            psp['rloc'] = float(rloc)  # Range of local Gaussian charge distribution
            for i, val in enumerate(fh.readline().split()):
                cloc[i] = float(val)
            psp['cloc'] = cloc  # Coefficients for the local part
            lmax = int(fh.readline().split()[0])
            psp['lmax'] = lmax  # Maximal angular momentum in the non-local part
            for iiter in range(lmax):
                rp[iiter], Nproj_l[iiter] = [float(i) for i in fh.readline().split()]
                for jiter in range(Nproj_l[iiter]):
                    tmp = fh.readline().split()
                    for iread, kiter in enumerate(range(jiter, Nproj_l[iiter])):
                        h[iiter, jiter, kiter] = float(tmp[iread])
            psp['rp'] = rp  # Projector radius for each angular momentum
            psp['Nproj_l'] = Nproj_l  # Number of non-local projectors
            for k in range(3):
                for i in range(2):
                    for j in range(i + 1, 2):
                        h[k, j, i] = h[k, i, j]
            psp['h'] = h  # Projector coupling coefficients per AM channel
    except FileNotFoundError:
        log.exception(f'There is no GTH pseudopotential for "{os.path.basename(f_psp)}"')
        raise
    return psp



#!/usr/bin/env python3
'''Package version number and version info function.'''
import os
import sys

__version__ = '2.0.0'
dependencies = ('numpy', 'scipy')
addons = ('vispy', 'nglview', 'notebook', 'jupyter_rfb', 'pylibxc', 'pyflosic2')
dev = ('pytest', 'flake8', 'sphinx', 'furo')


def info():
    '''Print version numbers and availability of packages.'''
    print('--- Version infos ---')
    print(f'python      : {sys.version.split()[0]}')
    print(f'eminus      : {__version__}')
    for pkg in dependencies + addons + dev:
        try:
            exec(f'import {pkg}')
            try:
                print(f'{pkg.ljust(12)}: {eval(pkg).__version__}')
            except AttributeError:
                # PyLibXC does not use the standard version identifier
                print(f'{pkg.ljust(12)}: {eval(pkg).version.__version__}')
        except ModuleNotFoundError:
            if pkg in dependencies:
                print(f'{pkg.ljust(12)}: Dependency not installed')
            elif pkg in addons:
                print(f'{pkg.ljust(12)}: Addon not installed')

    print('\n--- Performance infos ---')
    try:
        THREADS = int(os.environ['OMP_NUM_THREADS'])
    except KeyError:
        THREADS = 1
        print('INFO: No OMP_NUM_THREADS environment variable was found.\n'
              'To improve performance, add "export OMP_NUM_THREADS=THREADS" to your ".bashrc".\n'
              'Make sure to replace "THREADS", typically with the number of cores your CPU has.')
    finally:
        print(f'FFT operations will run on {THREADS} thread{"s" if THREADS!=1 else ""}.')
    return


if __name__ == '__main__':
    info()



#!/usr/bin/env python3
'''Setup file to make eminus installable.

For a full list of options see the documentation:
https://setuptools.pypa.io/en/latest/references/keywords.html
'''
from setuptools import find_packages, setup

with open('eminus/version.py', 'r') as fh:
    version = {}
    exec(fh.read(), version)

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='eminus',
    version=version['__version__'],
    description='A plane wave density functional theory code.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Wanja Timm Schulze',
    author_email='wangenau@protonmail.com',
    url='https://nextdft.gitlab.io/eminus',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Education',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Software Development'
    ],
    license='APACHE2.0',
    install_requires=['numpy', 'scipy'],
    keywords=['NextDFT'],
    package_data={
        'eminus.pade_gth': ['*.gth']
    },
    zip_safe=False,
    extras_require={
        'addons': ['jupyter_rfb', 'nglview', 'notebook', 'pyflosic2', 'pylibxc2', 'vispy'],
        'dev': ['flake8', 'flake8-docstrings', 'flake8-import-order', 'furo', 'pytest', 'sphinx']
    },
    python_requires='>=3.6',
    project_urls={
        'Documentation': 'https://nextdft.gitlab.io/eminus',
        'Source': 'https://gitlab.com/nextdft/eminus',
        'Bug Tracker': 'https://gitlab.com/nextdft/eminus/-/issues'
    }
)



#!/usr/bin/env python3
'''Fermi-orbital descriptor generation.'''
import os

import numpy as np
from numpy.linalg import norm
try:
    from pyflosic2.atoms.atoms import Atoms
    from pyflosic2.guess.pycom import pycom
    from pyflosic2.parameters.flosic_parameters import parameters
    from pyscf.gto import M  # PySCF is a dependency of PyFLOSIC2
    from pyscf.scf import RKS
except ImportError:
    print('ERROR: Necessary addon dependencies not found. To use this module,\n'
          '       install the package with addons, e.g., with "pip install eminus[addons]"')

from ..data import symbol2number
from ..filehandler import read_xyz
from ..units import bohr2ang


def get_fods(object, basis='pc-0', loc='FB', clean=True):
    '''Generate FOD positions using the PyCOM method.

    Reference: J. Comput. Chem. 40, 2843.

    Args:
        object: Atoms or SCF object.

    Keyword Args:
        basis (str): Basis set for the DFT calculation.
        loc (str): Localization method (case insensitive).
        clean (bool): Remove log files.

    Returns:
        ndarray: FOD positions.
    '''
    try:
        atoms = object.atoms
    except AttributeError:
        atoms = object
    loc = loc.upper()

    # Convert to Angstrom for PySCF
    X = bohr2ang(atoms.X)
    # Build the PySCF input format
    atom_pyscf = list(zip(atoms.atom, X))

    # Do the PySCF DFT calculation
    spin = np.sum(atoms.Z) % 2
    mol = M(atom=atom_pyscf, basis=basis, spin=spin)
    mf = RKS(mol=mol)
    mf.verbose = 0
    mf.kernel()

    # Add some FODs to the positions, otherwise the method will not work
    extra = np.zeros((atoms.Ns, 3))
    atom_pyflosic = atoms.atom + ['X'] * len(extra)
    X_pyflosic = np.vstack((X, extra))

    # Do the pycom FOD generation
    atoms = Atoms(atom_pyflosic, X_pyflosic, elec_symbols=['X', None], spin=spin)
    p = parameters(mode='restricted')
    p.init_atoms(atoms)
    p.basis = basis
    p.pycom_loc = loc
    pc = pycom(mf=mf, p=p)
    pc.get_guess()

    # Get the actual FOD positions from the xyz file
    atom, X = read_xyz(f'{loc}_GUESS_COM.xyz')
    _, _, fods = split_fods(atom, X)

    if clean:
        os.remove(p.log_name)
        os.remove(f'{loc}_GUESS_COM.xyz')
    return fods


def split_fods(atom, X):
    '''Split atom and FOD coordinates.

    Args:
        atom (list): Atom symbols.
        X (ndarray): Atom positions.

    Returns:
        tuple[list, ndarray, ndarray]: Atom types, the respective coordinates, and FOD positions.
    '''
    X_fod = []
    # Iterate in reverted order, because we may delete elements
    for ia in range(len(X) - 1, -1, -1):
        if atom[ia] == 'X':
            X_fod.append(X[ia])
            X = np.delete(X, ia, axis=0)
            del atom[ia]
    X_fod = np.asarray(X_fod)
    return atom, X, X_fod


def remove_core_fods(object, fods):
    '''Remove core FODs from a set of FOD coordinates.

    Args:
        object: Atoms or SCF object.
        fods (ndarray): FOD positions.

    Returns:
        ndarray: Valence FOD positions.
    '''
    try:
        atoms = object.atoms
    except AttributeError:
        atoms = object

    # If the number of valence electrons is the same as the number of FODs, do nothing
    if len(fods) == np.sum(atoms.Z) // 2:
        return fods

    for ia in range(atoms.Natoms):
        n_core = symbol2number[atoms.atom[ia]] - atoms.Z[ia]
        # In the spin-paired case two electrons are one state
        # Since only core states are removed in pseudopotentials this value is divisible by 2
        # +1 to account for uneven amount of core FODs (e.g., for hydrogen)
        n_core = (n_core + 1) // 2
        dist = norm(fods - atoms.X[ia], axis=1)
        idx = np.argsort(dist)
        # Remove core FODs with the smallest distance to the core
        fods = np.delete(fods, idx[:n_core], axis=0)
    return fods
